﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Windows.Input;
using System.ComponentModel;

namespace SigmaFilterForRealImages
{ public partial class MainPage : UserControl
  { WriteableBitmap img0, img1;
    Int32 kw, hkw;  //kernel width  and half kernel width,  kw uneven and hkw even
    Int32 kh, hkh;  //kernel height and half kernel height, kh uneven and hkh even
    Int32 xSize, ySize; //image width and height
    Int32 sigma=0;
    Int64 t0, t1;       //system time before and after filtering
    String fileName; 
    Byte [,] R0, G0, B0; //input
    Int32[,] R1, G1, B1; //intermediary
    Int32[,] R2, G2, B2; //output
    BackgroundWorker myThread = new BackgroundWorker(); //Thread object

    public MainPage()
    { InitializeComponent();
      myThread.DoWork += new DoWorkEventHandler( thread_DoWork );
      myThread.RunWorkerCompleted += 
        new RunWorkerCompletedEventHandler( update_View_when_thread_completed );
      myThread.WorkerSupportsCancellation = true;
    }

    private void myImageIn_ImageOpened( object sender,RoutedEventArgs e ) 
    { img0  = new WriteableBitmap( (BitmapSource)myImageIn.Source );
      fileName = ((BitmapImage)myImageIn.Source).UriSource.OriginalString;
      a_new_image_opened();
    }

    private void myButtonOF_Click( object sender, RoutedEventArgs e )
    { OpenFileDialog openFileDialog = new OpenFileDialog();
      openFileDialog.Filter = "Image Files (*.png,*.jpg)|*.png;*.jpg";
      bool? userClickedOK = openFileDialog.ShowDialog();
      if ( userClickedOK == true )
      { try
        { System.IO.Stream stream = openFileDialog.File.OpenRead();
          fileName = openFileDialog.File.ToString();
          BitmapImage bmp = new BitmapImage();
          bmp.SetSource( stream );
          stream.Close();
          img0 = new WriteableBitmap( bmp );
          a_new_image_opened();
        } catch {}
      }
    }

    private void a_new_image_opened()
    { xSize = img0.PixelWidth;
      ySize = img0.PixelHeight;
      img1 = new WriteableBitmap( xSize, ySize );
      myImageIn .Source = img0;
      myImageOut.Source = img1;
      //prepare_intermediary_arrays
      R0 = new Byte [ySize, xSize];
      G0 = new Byte [ySize, xSize];
      B0 = new Byte [ySize, xSize];
      R1 = new Int32[ySize, xSize];
      G1 = new Int32[ySize, xSize];
      B1 = new Int32[ySize, xSize];
      R2 = new Int32[ySize, xSize];
      G2 = new Int32[ySize, xSize];
      B2 = new Int32[ySize, xSize];
      GC.Collect(); //clean up memory when new arrays override old ones
      for ( int y=0; y < ySize; y++ )
        for ( int x=0; x < xSize; x++ )
        { Int32 c = img0.Pixels[y*xSize+x];
          R0[y,x] = (Byte)( c & 0x0000ff)       ;
          G0[y,x] = (Byte)((c & 0x00ff00) >>  8);
          B0[y,x] = (Byte)((c & 0xff0000) >> 16);
        }
      width_Slider .Maximum = xSize;
      height_Slider.Maximum = ySize;
      if ( update_View_before_starting_thread() ) start_myThread();
    }

    private bool update_View_before_starting_thread()
    { myTextBlockOF.Text = fileName + ": " +
          xSize.ToString() + 'x' + ySize.ToString();
      sigma=(Int32)sigma_Slider .Value;
      kw   =(Int32)width_Slider .Value; if (kw%2==0) kw--; hkw=kw/2; //filter width
      kh   =(Int32)height_Slider.Value; if (kh%2==0) kh--; hkh=kh/2; //filter height
      int kernelArea = kw*kh; //no. of pixels inside the filter kernel
      myTextBlockSliders.Text = "FilterWidth  = " + kw.ToString() + ", " +
                                "FilterHeight = " + kh.ToString();
      for ( int i=0; i < xSize*ySize; )
        unchecked { img1.Pixels[i++] = (Int32)0xffffffff; }
      img1.Invalidate();
      myTextBlockTime.Text = "Wait !"; return true;
    }

    private void start_myThread()
    { if ( myThread.IsBusy ) return;
      myBorder.Cursor = Cursors.Wait;
      t0 = DateTime.Now.Ticks; 
      for ( int i=0; i < xSize*ySize; )   
        unchecked { img1.Pixels[i++] = (Int32)0xffffffff; }
      img1.Invalidate();
      myThread.RunWorkerAsync();
    }

    private void update_View_when_thread_completed( Object sender,	RunWorkerCompletedEventArgs e )
    { myBorder.Cursor = Cursors.Arrow;
      if ( e.Cancelled ) 
      { myTextBlockTime.Text = "";
        TextBox_TextChanged( null, null ); //try again
      }
      else 
      { t1 = DateTime.Now.Ticks;
        myTextBlockTime.Text =  ((t1-t0)/10000).ToString() + " millisec.";
        img1.Invalidate();
      }
    }

    private void Slider_MouseLeftButtonUp( object sender, MouseButtonEventArgs e ) 
    { sigmaTextBox.Text = ((Int32)sigma_Slider .Value).ToString();
      kwTextBox   .Text = ((Int32)width_Slider .Value).ToString();
      khTextBox   .Text = ((Int32)height_Slider.Value).ToString();
    }

    private void TextBox_TextChanged( object sender, TextChangedEventArgs e ) 
    { if ( myThread.IsBusy ) { myThread.CancelAsync(); return; }
      int width=0, height=0;
      try { sigma  = Convert.ToInt32( sigmaTextBox.Text ); } catch { return; }
      try { width  = Convert.ToInt32( kwTextBox   .Text ); } catch { return; }
      try { height = Convert.ToInt32( khTextBox   .Text ); } catch { return; }
      if ( sigma  < 0 | sigma  > 255   ) return;
      if ( width  < 1 | width  > xSize ) return;
      if ( height < 1 | height > ySize ) return;
      sigma_Slider .Value = sigma;
      width_Slider .Value = width;
      height_Slider.Value = height;
      if ( update_View_before_starting_thread() ) start_myThread();
    }

    private void thread_DoWork( Object sender, DoWorkEventArgs e )
    { sigmaFilter( e );
    }

    private void sigmaFilter( DoWorkEventArgs e )
    { int R, G, B, Rsum, Gsum, Bsum, x, xx, xxx, y, yy, yyy, startOfYRow;
      int RmidPixel, GmidPixel, BmidPixel, RanyPixel, GanyPixel, BanyPixel;
      int Rno, Gno, Bno;
      int kernelArea = kw*kh; //no of pixels inside the kernel
      for ( y=0; y < ySize; y++ ) //==================
      { if ( myThread.CancellationPending ) e.Cancel = true; //kill the thread
        startOfYRow = y*xSize;
        for ( x=0; x < xSize; x++ ) //===============
        { RmidPixel = R0[y,x];
          GmidPixel = G0[y,x];
          BmidPixel = B0[y,x];
          Rsum=Gsum=Bsum=Rno=Gno=Bno=0;
          for ( yy=-hkh; yy <= hkh; yy++ ) //=============
          { yyy = y + yy;
            if ( yyy < 0 | yyy >= ySize ) continue;   //beyond the border
            for ( xx=-hkw; xx <= hkw; xx++ )//========
            { xxx = x + xx;
              if ( xxx < 0 | xxx >= xSize ) continue; //beyond the border
              RanyPixel = R0[yyy,xxx];
              GanyPixel = G0[yyy,xxx];
              BanyPixel = B0[yyy,xxx];
              if (Math.Abs(RmidPixel-RanyPixel) <= sigma) { Rsum+=RanyPixel; Rno++; }
              if (Math.Abs(GmidPixel-GanyPixel) <= sigma) { Gsum+=GanyPixel; Gno++; }
              if (Math.Abs(BmidPixel-BanyPixel) <= sigma) { Bsum+=BanyPixel; Bno++; }
            } //====== end for (int xx... ================
          } //======== end for (int yy... ==================
          if (Rno != 0) { R=Convert.ToInt32((float)Rsum/Rno);       } else R = RmidPixel;
          if (Gno != 0) { G=Convert.ToInt32((float)Gsum/Gno) <<  8; } else G = GmidPixel;
          if (Bno != 0) { B=Convert.ToInt32((float)Bsum/Bno) << 16; } else B = BmidPixel;
          unchecked { img1.Pixels[startOfYRow+x] = (Int32)0xff000000 | R | G | B; }
        } //============ end for (int x... =====================
      } //============== end for (int y... =======================
    }
  }
}